(* ::Package:: *)

(* AutoPortrait *)
(* Automatic Generation of Interactive Multidimensional Phase Portraits *)
(* Copyright Lobo Lab (https://lobolab.umbc.edu) *)
(* GNU General Public License v3 *)

BeginPackage["AutoPortrait`"]

GeneratePortrait::usage="GeneratePortrait[dxs,xs] generates the phase portraits for the system of differential equations dxs with dependent variables xs.";
AutoPortraitGUI::usage="AutoPortraitGUI[] displays the graphical user interface to generate phase portraits.";

Begin["`Private`"]

GeneratePortrait[dxs_,xs_,OptionsPattern[{ParamValues->{},ParamRanges->{},PortraitRange->Null,Traj->Null,TrajRange->10,NonNegativeDomain->False,NumColumns->3,ShowLegend->True}]]:= DynamicModule[{J,ps,psv,psr,params,paramSliders,criticalPoints,CalcStability,CalcPlotRange,plotRange, stability, traj,t,sys,bound,tr,sol},
	traj=OptionValue[Traj];
	tr=OptionValue[TrajRange];
	If[traj=!=Null, bound = MapThread[(#1 == #2) &, {Construct[#, 0] & /@ xs, traj}]];
	
	If[OptionValue[NonNegativeDomain],
		sys=MapThread[(#1==Piecewise[{{#2,#3>=0}},0])&,{Construct[#',t]&/@xs, dxs/.MapThread[#1->#2&,{xs,Construct[#,t]&/@xs}], Construct[#,t]&/@xs}],
		sys=MapThread[(#1==#2)&,{Construct[#',t]&/@xs,dxs/.MapThread[#1->#2&,{xs,Construct[#,t]&/@xs}]}]
	];
	
	(* Parameters *)
	ps=DeleteCases[DeleteDuplicates@Cases[dxs,_Symbol,Infinity],Alternatives@@xs];
	psv=Join[OptionValue[ParamValues],#->1&/@DeleteCases[ps,Alternatives@@Keys[OptionValue[ParamValues]]]];
	psr=Join[OptionValue[ParamRanges],#->{0,2*#/.psv}&/@DeleteCases[ps,Alternatives@@Keys[OptionValue[ParamRanges]]]];

	(* Display numbers in variables and parameters as subscripts *)
	MapThread[(Format[#1]:= Style[#2])&, {Join[xs,ps], 
   StringJoin[
      Function[{$s}, 
        If[StringMatchQ[$s, DigitCharacter ..], 
         ToString[Subscript["", $s], StandardForm], $s]] /@ 
       StringCases[ToString[#], 
        DigitCharacter .. | LetterCharacter ..]]& /@ Join[xs,ps]}];

	CalcStability[criticalPoint_]:=Module[{Jcp,EigVals, stab},
		Jcp = J/.criticalPoint;
		EigVals= Eigenvalues[Jcp];

		Which[Max[Re[EigVals]]<0,stab=Green, (* Stable point *)
			Min[Re[EigVals]]>0,stab=Red, (* Unstable point *)
			ContainsAny[Re[EigVals],{0}],stab=Black, (* Nonhyperbolic point: we don't know *)
			True,stab=Yellow]; (* Saddle point *)

		Return[stab];
	];

	CalcPlotRange[]:=Module[{range, rangePad},
		range=MinMax[xs/.criticalPoints/.(#->0&/@xs)];
		rangePad=0.5*(range[[2]]-range[[1]]);
		If[rangePad==0,rangePad=1];
		Return[Table[range+If[OptionValue[NonNegativeDomain],{0,rangePad},{-rangePad,rangePad}],2]];
	];

	criticalPoints=NSolve[(dxs/.psv)==0,xs,If[OptionValue[NonNegativeDomain],NonNegativeReals,Reals]];
	J=D[dxs/.psv,{xs}];
	stability=CalcStability/@criticalPoints;
	plotRange=OptionValue[PortraitRange];
	If[OptionValue[PortraitRange]===Null, plotRange=CalcPlotRange[],plotRange=Table[OptionValue[PortraitRange],2]];

	Column[{
		(* Parameter sliders *)
		If[Length[ps]>0,(
			paramSliders=ps/.psv;

			Manipulate[
				params=MapThread[#1->#2 &,{ps,paramSliders}];

				criticalPoints=NSolve[(dxs/.params)==0,xs,If[OptionValue[NonNegativeDomain],NonNegativeReals,Reals]];
				J=D[dxs/.params,{xs}];
				stability=CalcStability/@criticalPoints;
				plotRange=CalcPlotRange[];

				Grid[Table[With[{$k=$k},{
					Style[ps[[$k]],Italic], ": ",
					Slider[Dynamic[paramSliders[[$k]]],{#[[1]],#[[2]],0.01*(#[[2]]-#[[1]])}&/@ps[[$k]]/.psr,ImageSize->300],
					InputField[Dynamic[paramSliders[[$k]]],Number,FieldSize->{{1,4},1}]
				}],{$k,Length[ps]}]],

				FrameLabel->{{None,None},{None,Style["Parameters",18]}},
				SynchronousUpdating->False,TrackedSymbols:>{paramSliders}
			]),(
			params={};
			paramSliders={};
			""
		)],

		Multicolumn[Flatten[{
			Table[With[{$i= $i},
				Table[With[{$j=$j},
					DynamicModule[{othervs,others,rangeh,rangev,otherSliders,viewRange,mouseDragInit},

						othervs=Delete[xs,{{$i},{$j}}];
						otherSliders=If[Length[criticalPoints]>0, othervs/.First[criticalPoints[[Ordering[Count[Chop[#],0]&/@Values[criticalPoints],1]]]]/.(#->0&/@xs),Table[0,Length[othervs]]];
						viewRange=plotRange;

						Manipulate[
							rangeh={xs[[$i]] ,viewRange[[1,1]], viewRange[[1,2]]};
							rangev={xs[[$j]] ,viewRange[[2,1]], viewRange[[2,2]]};
							others=MapThread[#1->#2 &,{othervs,otherSliders}];

							EventHandler[Show[{ 
								ContourPlot[{(dxs[[$i]]/.params/.others)==0,(dxs[[$j]]/.params/.others)==0},rangeh+{0,-0.1,0.1}, rangev+{0,-0.1,0.1}, ContourStyle->Purple, PlotPoints-> 150],

								StreamPlot[dxs[[{$i,$j}]]/.params/.others,rangeh, rangev,StreamPoints->Automatic,StreamColorFunction->None,PerformanceGoal->"Quality"],

								If[traj===Null,
									Graphics[],
									ParametricPlot[Evaluate[{xs[[$i]][t],xs[[$j]] [t]}/.sol],{t,0,tr},PlotStyle->{Thickness[0.004],Red}]/. Line[$x_]->{
										Arrowheads[{0.04,0}],
										(*If[otherSliders!=Delete[traj,{{$i},{$j}}],Opacity[0.5]],*)
										Tooltip[EventHandler[Arrow[$x],{"MouseClicked",1}:>(If[CurrentValue["ControlKey"],otherSliders=othervs/.AssociationThread[xs->traj]])],traj]
									}],

								If[Length[criticalPoints]>0,
									Table[With[{$k=$k},{
									
										(* Check if the critical point contains other variables *)
										If[Length[others]>0&&Or@@Table[With[{$l=$l},MemberQ[criticalPoints[[$k]], Keys[others][[$l]],Infinity]],{$l, Length[others]}],
											If[MemberQ[Keys[criticalPoints[[$k]]], xs[[$i]], Infinity] && MemberQ[Keys[criticalPoints[[$k]]], xs[[$j]], Infinity],
												Graphics[{
													PointSize[Large],
													If[otherSliders!=(othervs/.criticalPoints[[$k]]), Opacity[0.75]], 
													stability[[$k]], 
													Tooltip[EventHandler[Point[{xs[[$i]],xs[[$j]]}]/. criticalPoints[[$k]],{"MouseClicked",1}:>(If[CurrentValue["ControlKey"],otherSliders=(othervs/.criticalPoints[[$k]])])],criticalPoints[[$k]]]
												}],
											
												Graphics[{
													Opacity[0.5],
													Orange,
													Rectangle[{viewRange[[1,1]],viewRange[[2,1]]},{viewRange[[1,2]],viewRange[[2,2]]}]}]
											],

											(* Check if critical point contain both xs[[i]] and xs[[j]]; if so plot them as a point *)
											If[MemberQ[Keys[criticalPoints[[$k]]], xs[[$i]], Infinity] && MemberQ[Keys[criticalPoints[[$k]]],xs[[$j]],Infinity],
												Graphics[{
													PointSize[Large], 
													stability[[$k]], 
													Tooltip[Point[{xs[[$i]], xs[[$j]]}]/.criticalPoints[[$k]],criticalPoints[[$k]]]}],

												(* Check if critical point is either xs[[i]] or xs[[j]] and plot the expression with respect to the other variables in the function *)
												If[MemberQ[Keys[criticalPoints[[$k]]], xs[[$i]],Infinity] || MemberQ[Keys[criticalPoints[[$k]]], xs[[$j]],Infinity],
													With[{$x=xs[[$j]]},
														ParametricPlot[Tooltip[{xs[[$i]]/.criticalPoints[[$k]],xs[[$j]]/.criticalPoints[[$k]]},criticalPoints[[$k]]], {$x, viewRange[[1,1]], viewRange[[1,2]]}, PlotRange->viewRange, PlotStyle->Orange]],

												Graphics[]
												]
											]
									]}],{$k, Length[criticalPoints]}],

									Graphics[]
								]}, 

								PlotRange->Dynamic[viewRange],FrameLabel->{{None,Style[xs[[$j]],Italic]},{None,Style[xs[[$i]],Italic]}},RotateLabel->False, Axes->True,FrameStyle->Directive[Black,FontColor->Black],BaseStyle->Directive[FontSize->18],Method->{"AxesInFront"->False}
								],{
	
								(* Mouse interactive controls *)
								
								{"MouseDown", 1} :> (
									mouseDragInit = MousePosition["Graphics"];
									),
									
								{"MouseDragged", 1} :> (
									viewRange-=(MousePosition["Graphics"] - mouseDragInit);
									If[OptionValue[NonNegativeDomain],viewRange-={Min[0,viewRange[[1,1]]],Min[0,viewRange[[2,1]]]}];
									),
									
								{"MouseClicked", 1} :> (
									If[!CurrentValue["ControlKey"],
										If[CurrentValue["ShiftKey"],{
											traj=Flatten[xs/.others/.MapThread[#1->#2&,{xs[[{$i,$j}]],MousePosition["Graphics"]}]];
											bound=MapThread[(#1==#2)&,{Construct[#,0]&/@xs,traj}];
											sol=NDSolve[Join[sys/.params,bound],xs,{t,tr}];
										},{
											viewRange+=MousePosition["Graphics"]-Mean/@viewRange;
											viewRange+=1/4*{{-1,1},{-1,1}}*(Subtract@@@viewRange/2);
											If[OptionValue[NonNegativeDomain],viewRange=Clip[viewRange,{0,Infinity}]];
									}]]),
								
								{"MouseClicked",2} :> (
									If[CurrentValue["ControlKey"],
										viewRange=plotRange,
										If[CurrentValue["ShiftKey"],{
												traj=Null;
											},{
												viewRange+=1/3*{{1,-1},{1,-1}}*(Subtract@@@viewRange/2);
												If[OptionValue[NonNegativeDomain],viewRange=Clip[viewRange,{0,Infinity}]];
											}]
									])
								}
							],

							(* Other variable sliders *)
							Grid[Table[With[{$k=$k},{
									Style[othervs[[$k]],Italic], ": ",
									Slider[Dynamic[otherSliders[[$k]]],{plotRange[[1,1]],plotRange[[1,2]],0.01*(plotRange[[1,2]]-plotRange[[1,1]])},ImageSize->320],
									InputField[Dynamic[N[otherSliders[[$k]],10]],Number,FieldSize->{{1,4},1}]
								}],{$k,Length[othervs]}],ItemStyle->Directive[FontSize -> 12]],

							FrameLabel->{{None,None},{None,Style[Row[{"Portrait ", Style[xs[[$i]],Italic], " - ", Style[xs[[$j]],Italic]}],18]}},
							SynchronousUpdating->False,TrackedSymbols:>{otherSliders,plotRange,viewRange,traj,sol,tr,stability,criticalPoints,params}
						]
					]],
				{$j,$i+1,Length[xs]}]],
			{$i,Length[xs]}],

			(* Legend *)
			If[OptionValue[ShowLegend],
				Style[Grid[{{
					Column[{
						Style[
							"Legend:",12],
							PointLegend[{ColorData[97][1],Red},{"Stream lines","Trajectory"},LegendMarkers->Graphics[{Arrowheads[0.4],Arrow[{{0,0},{5,0}}]}],LegendMarkerSize->20],
							LineLegend[{Purple,Orange},{"Nullcline","Critical line"},LegendMarkerSize->20],
							PointLegend[{Green,Red, Yellow, Black},{"Stable critical point","Unstable critical point","Saddle critical point", "Nonhyperbolic critical point"},LegendMarkerSize->13],
							"",
							Style["Critical points not in the plane",10],
							Style["are displayed semitransparent.",10]
					}],
					Column[{
						Style[
							"Portrait interactive controls:",12],
							PointLegend[Table[Black,7],{
							"Left click: zoom in",
							"Right click: zoom out",
							"Left click drag: pan",
							"Left click + Ctrl on element: center",
							"Right click + Ctrl: restore default view",
							"Left click + Shift: add trajectory",
							"Right click + Shift: remove trajectory"},LegendMarkers->"-"]
					}]
				}},Alignment->Top, Spacings->{2,2},Frame-> All,FrameStyle->Directive[Gray, Thickness@0.5]],FontFamily->"Helvetica",10.5]
			],

			(* Trajectory plot *)
			Dynamic[If[traj=!=Null,
				Manipulate[
					sol=NDSolve[Join[sys/.params,bound],xs,{t,tr}];
					Show[{
						If[traj===Null,
							Graphics[],
							Plot[Evaluate[Construct[#,t]&/@xs/.sol],{t,0,tr},PlotRange->Full,PlotLegends->Placed[LineLegend[xs, LabelStyle->Directive[Italic,FontSize->18]],Above],Frame->True,FrameStyle->Directive[Black,FontColor->Black],FrameLabel->{"Time",None},LabelStyle->Directive[FontSize->18],AspectRatio->0.83]]
						},BaseStyle->{FontSize->18}
					],
					Row[{"Time period = ",
					InputField[Dynamic[tr],Number,ImageSize->50]},BaseStyle->12],
					FrameLabel->{{None,None},{None,Style["Trajectory",18]}},
					SynchronousUpdating->False,TrackedSymbols:>{traj,sol,tr,params,bound}
				],
				""]]

	}],OptionValue[NumColumns],Appearance->"Horizontal"]}]
];


(* AutoPortrait GUI *)
AutoPortraitGUI[OptionsPattern[{NumColumns->3,ShowLegend->True}]]:=DynamicModule[{n=3,variables=Table[Symbol["x" <> ToString@i], {i, 10}],equations=Table[Symbol["x" <> ToString@i], {i, 10}], ps={},pValues={},pRanges={},FindParams, nplots=0, nonnegative=False,newVar,newFun,portrait=""},

	ClearAll["Global`*"];
	
	FindParams[]:=Module[{psv=MapThread[#1->#2&,{ps,pValues}],psr=MapThread[#1->#2&,{ps,pRanges}]},
		ps=DeleteCases[DeleteDuplicates@Cases[Take[equations,n],_Symbol,Infinity],Alternatives@@Take[variables,n]];
		pValues=ps/.psv/.(#->1&/@ps);
		pRanges=ps/.psr/.(#->{0,2}&/@ps);
	];

	Column[{
		Style["AutoPortrait",Bold,24],
		Style["Automatic Phase Portrait Generator",20],
		Style["Lobo Lab",16],
		"",
		Grid[{
				{"Number of equations: ", Dynamic[n], " ", Slider[Dynamic[n,(portrait="";n=#;FindParams[])&],{2,10,1},ImageSize->340]},
				{"Nonnegative domain: ",Checkbox[Dynamic[nonnegative,(portrait="";nonnegative=#)&]]}
			},Alignment->Right
		],
		Dynamic[Grid[{{
			Column[{
					"Differential Equations:",
					Grid[Table[With[{$i=$i},{ 
						OverDot[InputField[Dynamic[InputForm[variables[[$i]]],(variables[[$i]]=Cases[{#},_Symbol,\[Infinity]][[1]];FindParams[];portrait="")&],FieldSize->{{1,40},Small}]],
						"=",
						InputField[Dynamic[InputForm[equations[[$i]]],(equations[[$i]]=#;FindParams[];portrait="")&],FieldSize->{{1,40},Small}]
						}],{$i,n}], Alignment->Left
					]
				},Center, Frame-> True,FrameStyle->Directive[Thickness@0.5]
			],

			If[Length[ps]>0,
				Column[{
					"Parameters:",
					Grid[Table[With[{$i=$i},{ 
					ps[[$i]],"=",InputField[Dynamic[pValues[[$i]]],Number,FieldSize->{{1,80},Small}],
					"  ",
					ps[[$i]],"\[Element]","[",InputField[Dynamic[pRanges[[$i,1]]],Number,FieldSize->{{1,40},Small}],",",InputField[Dynamic[pRanges[[$i,2]]],Number,FieldSize->{{1,40},Small}],"]"
					}],{$i,Length[ps]}], Alignment->Left]
					},Center, Frame-> True,FrameStyle->Directive[Thickness@0.5]
				],
				""
			]}},Alignment->Top
		]],

		Button[Style["Generate portraits",12],
			portrait=GeneratePortrait[Take[equations,n],Take[variables,n],ParamValues->MapThread[#1->#2 &,{ps,pValues}],ParamRanges->MapThread[#1->#2 &,{ps,pRanges}],NonNegativeDomain->nonnegative,NumColumns->OptionValue[NumColumns],ShowLegend->OptionValue[ShowLegend]]
		],
		Dynamic[portrait]
	},Center, Frame-> True,FrameStyle->Directive[Thickness@1]]
];


End[]
EndPackage[]
